#!/usr/bin/env python
#-*- coding: ISO-8859-1 -*-
#
# Author: Brain Bockelman

import FileManager
import DBSInteraction
import EdmFileExtraction

class EdmFileManager(ConfiguredObjects):

    def __init__(self):
        self.section = "edm_file_manager"
        self._server = FileManager.Server
        self._dbs = DBSInteraction.DBS
        self._run_cache = {}
        self._running_extractions = {}
        self._extractions_status = {}

    def configure(self, cp):
        self._server.configure(cp)
        self.pool = FileManager.ThreadPool(threads=int(self.getOption( \
            "max_extractions", "5")))
        self.lfn_src = self._server.getOption("base_directory", "/var/www/html")
        self.run_dest = self.getOption("base_directory",
            "/var/www/html/extracted")

    def getEvent(self, run, event):
        raise NotImplementedError()
        # This will be called after edmFileUtil has extracted the event.
        # (status returns OK).  
        
    def _getLfnsForRun(self, run):
        if run not in self._run_cache:
            self._run_cache[run] = self._dbs.getFilesForRun(run)
        return self._run_cache[run]

    def _handleRunningExtractions():
        failed_keys = []
        self._extraction_lock.acquire()
        try:
            for key, extraction in self._running_extractions.items():
                self._extractions_status[key] = extraction.status()
                if status[0] == 3: # FAILED!
                    failed_keys.append(key)
            for key in failed_keys:
                del self._running_extractions[key]
        finally:
            self._extraction_lock.release()

    def _launchExtraction(self, run, event):
        key = run, event
        if key in self._running_extractions:
            self.log.warning("Internal error: about to launch an extraction " \
                "for something which is already running! Run %s, event %s" % \
                (run, event))
        extracter = EdmFileExtraction(self.cp)
        extracter.request(run, event, self._getLfnsForRun(run), self.lfn_src,
            self.run_dest)
        self.pool.queue(extracter)
        self._running_extractions(key) = extracter

    def status(self, run, event):
        if run, event in self._running_extractions:
            self._handleRunningExtractions()
        if run, event in self._extractions_status:
            return self._extractions_status[run, event]
        lfns = self._getLfnsForRun(run)
        worst_status = 0
        status_messages = []
        for lfn in lfns:
            status_code, status_msg = self._server.status(lfn)
            if status_code > worst_status:
                worst_status = status_code
            status_messages.append("LFN %s status message: %s" % (lfn,
                status_msg))
        if worst_status > 0:
            return worst_status, '\n'.join(status_messages)
        
        # First, check with the server and report the "worst" file status
        # If all the files are working, return "extraction in progress"
        # This is what launches the edmFileUtil in a separate thread.
        # Once edmFileUtil has successfully found our event, then we
        # return the success code.

    def request(self, run, event, user=None):
        lfns = self._getLfnsForRun(run)
        for lfn in lfns:
            self._server.request(lfns, user=user)

    def cancel(self, run, event, user=None):
        for lfn in self._getLfnsForRun(run):
            self._server.cancel(lfn, user=user)

    def graceful_exit(self):
        self._server.graceful_exit()

