
import os
import errno
import threading

from fuse import Fuse
from pyGlobus.ftpClient import FtpClient, HandleAttr, OperationAttr, \
     DebugPlugin, RestartMarkerPlugin, RestartMarker
from pyGlobus.util import Buffer, GlobusException

from FileManager import ThreadPool, ConfiguredObject

class BlockRequest(object):

    def __init__(self, offset, requested):
        self.offset, self.requested = offset, requested

    def partial_block_can_satisfy(self, partial_block):
        if partial_block.offset <= self.offset and partial_block.offset + \
                partial_block.current >= self.offset + self.requested:
            return True
        return False

    def transfer_can_satisfy(self, transfer):
        if transfer.current_marker >= self.offset + self.requested:
            return True
 
class PartialBlock(object):

    def __init__(self, offset, requested, parent_file, data=None):
        self.offset, self.requested = offset, requested
        self.parent_file = parent_file
        if data:
            assert len(data) == requested
            self.complete = True
            self.current = requested
        else:
            self.current = 0
            self.complete = False

    def fulfill_request(self, block_request):
        my_start = block_request.offset - self.offset
        my_finish = block_request.offset + block_request.requested - self.offset
        return self.data[my_start:my_finish]

    def merge(self, partial_block):
        assert self.complete == True and partial_block.complete == True
        if partial_block.offset == self.offset:
            if partial_block.requested >= self.requested:
                return partial_block
            else:
                return self
        if partial_block.offset < self.offset:
            return partial_block.merge(self)
        # self.offset < partial_block.offset
        if self.offset + self.requested >= partial_block.offset:
            if self.offset + self.requested >= partial_block.offset + \
                    partial_block.requested:
                return self
            keep_len = self.offset + self.requested - partial_block.offset
            assert keep_len > 0 and keep_len < self.requested
            new_data = data[:keep_len] + partial_block.data
            new_size = keep_len + partial_bock.requested
            assert new_size == len(new_data)
            return PartialBlock(self.offset, new_size, self.parent_file,
                data=new_data)
        # Blocks do not overlap or touch.
        return None

    def finish(self):
        base_dir = self.partial_file
        self.complete = True
        filename = "%i_%i" % self.offset, self.request
        fp = open(os.path.join(base_dir, filename), "w")
        fp.write(self.data)
        fp.close()

class PartialFile(ConfiguredObject):

    def __init__(self, filename):
        self.blocks = []
        self.min_read = 10*1024*1024
        self.filename = filename

    def configure(self, cp):
        self.cp = cp

    def calculate_partial_block(self, block_request):
        # TODO: Make sure we don't overlap with any current partial block!
        partial_block = PartialBlock(block_request.offset, min(block_request.\
            request, self.min_read)
        self.blocks.append(partial_block)

class GridftpPartialRunner(object):

    def __init__(self, block_request, partial_file):
        self.block = partial_file.calculate_partial_block(block_request)
        self.block_request = block_request
        self.cb_handle = None
        self.buffer = Buffer(256*1024)
        self.handleAttr = HandleAttr()
        self._ftpClient = FtpClient(handleAttr)
        self._condV = threading.Condition(threading.Lock())
        self._condExternal = threading.Condition(threading.Lock())
        if DEBUG_LEVEL:
            try:
                debug_plugin = DebugPlugin(sys.stdout, "Get")
            except PluginException, ex:
                print "NO DEBUG PLUGIN"
            else:
                handleAttr.add_plugin(debug_plugin)
        try:
            plugin = RestartMarkerPlugin(self.beginCB, self.markerCB,
                self.completeCB)
        except PluginException, ex:
            if DEBUG_LEVEL:
                print "NO RESTART PLUGIN"
            else:
                pass #handleAttr.add_plugin(plugin)

    def wait_on_data(self):
        self._condExternal.acquire()
        self._condExternal.wait()
        self._condExternal.release()
        if self._done_success:
            return self.block.fulfill_request(self.block_request)

    def read_cb(self, handle, buffer, bufHandle, bufLen, offset, eof,
            error):
        if error:
            raise Exception("GridFTP error: %s" % error)
        self.block.write(buffer)
        if self.block_request.partial_block_can_satisfy()
            self._condExternal.acquire()
            self._condExternal.notify()
            self._condExternal.release()

    def done_func(self, cv, handle, error):
         if error or (error != "NONE"):
             print "The following error occured: %s" %(error)
         else:
             print "No Errors"
         cv.acquire()
         cv.notify()
         cv.release()

    def start(self):
         try:
             operationAttr = OperationAttr()
             restartMarker = RestartMarker()
             self._ftpClient.get(src_url, operationAttr, restartMarker,
                 self.block.offset, self.block.offset + self.block.requested,
                 self.done_func, self._condV)
             self.handle = self._ftpClient.register_read(self._buffer,
                 self.read_cb)
         except GlobusException, ex:
             raise

         self._condV.acquire()
         self._condV.wait()
         self._condV.release()
         fd.close()
         self._ftpClient.free_callback(handle)

    def cancel(self):
        self._ftpClient.abort()

class GridftpFUSE(Fuse, ConfiguredObject):

    def __init__(self):
        self.cp = None
        self.section = "gridftp_fuse"
        Fuse.__init__(self)
        ConfiguredObject.__init__(self)
        self._lock = threading.lock()
        self._files = {}

    def configure(self, cp):
        self.cp = cp
        self.transfer_pool_size = int(self.getOption("transfer_pool_size", "5"))
        self.partial_pool_size = int(self.getOption("partial_pool_size", "5"))
        self.transfer_pool = ThreadPool(threads=self.transfer_pool_size)
        self.partial_pool = ThreadPool(threads=self.partial_pool_size)

    def __init__(self, *args, **kw):
        Fuse.__init__(self, *args, **kw)

        print 'Init complete.'

    def getattr(self, path):
        """
        - st_mode (protection bits)
        - st_ino (inode number)
        - st_dev (device)
        - st_nlink (number of hard links)
        - st_uid (user ID of owner)
        - st_gid (group ID of owner)
        - st_size (size of file, in bytes)
        - st_atime (time of most recent access)
        - st_mtime (time of most recent content modification)
        - st_ctime (platform dependent; time of most recent metadata change on Unix,
                    or the time of creation on Windows).
        """

        print '*** getattr', path

        depth = getDepth(path) # depth of path, zero-based from root
        pathparts = getParts(path) # the actual parts of the path

        return -errno.ENOSYS


    def getdir(self, path):
        """
        return: [('file1', 0), ('file2', 0), ... ]
        """

        print '*** getdir', path
        return -errno.ENOSYS

    def mythread ( self ):
        print '*** mythread'
        return -errno.ENOSYS

    def chmod ( self, path, mode ):
        print '*** chmod', path, oct(mode)
        return -errno.ENOSYS

    def chown ( self, path, uid, gid ):
        print '*** chown', path, uid, gid
        return -errno.ENOSYS

    def fsync ( self, path, isFsyncFile ):
        print '*** fsync', path, isFsyncFile
        return -errno.ENOSYS

    def link ( self, targetPath, linkPath ):
        print '*** link', targetPath, linkPath
        return -errno.ENOSYS

    def mkdir ( self, path, mode ):
        print '*** mkdir', path, oct(mode)
        return -errno.ENOSYS

    def mknod ( self, path, mode, dev ):
        print '*** mknod', path, oct(mode), dev
        return -errno.ENOSYS

    def open ( self, path, flags ):
        print '*** open', path, flags
        return -errno.ENOSYS

    def read ( self, path, length, offset ):
        self.dbs.getlfn(path)
        self._lock.acquire()
        try:
            if path in self.files:
                partial_file = self.files[path]
            else:
                partial_file = self.files.setdefault(path, PartialFile())
        finally:
            self._lock.release()
        block_request = BlockRequest(offset, length)
        partial_block = partial_file.calculate_partial_block(block_request)
        gridftp = GridftpPartialRunner(partial_block, block_request)
        return gridftp.wait_on_data()

    def readlink ( self, path ):
        print '*** readlink', path
        return -errno.ENOSYS

    def release ( self, path, flags ):
        print '*** release', path, flags
        return -errno.ENOSYS

    def rename ( self, oldPath, newPath ):
        print '*** rename', oldPath, newPath
        return -errno.ENOSYS

    def rmdir ( self, path ):
        print '*** rmdir', path
        return -errno.ENOSYS

    def statfs ( self ):
        print '*** statfs'
        return -errno.ENOSYS

    def symlink ( self, targetPath, linkPath ):
        print '*** symlink', targetPath, linkPath
        return -errno.ENOSYS

    def truncate ( self, path, size ):
        print '*** truncate', path, size
        return -errno.ENOSYS

    def unlink ( self, path ):
        print '*** unlink', path
        return -errno.ENOSYS

    def utime ( self, path, times ):
        print '*** utime', path, times
        return -errno.ENOSYS

    def write ( self, path, buf, offset ):
        print '*** write', path, buf, offset
        return -errno.ENOSYS

if __name__ == '__main__':
    fs = GridftpFUSE()
    fs.flags = 0
    fs.multithreaded = 1
    fs.main()

